/**
 * File for image handling logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "image.h"
    #include "utility.h"
#endif

#ifndef user_headers_logging
    #define user_headers_logging
    #include "logging.h"
#endif


// Global Variables.
#ifndef LOGGING_CLASS_INIT
    #define LOGGING_CLASS_INIT
    extern Logging logger;
#endif
const Image::PixelStruct Image::pixelBlack = Image::PixelStruct(0);
const Image::PixelStruct Image::pixelWhite = Image::PixelStruct(1);
const Image::PixelStruct Image::pixelRed = Image::PixelStruct(1,0,0);
const Image::PixelStruct Image::pixelGreen = Image::PixelStruct(0,1,0);
const Image::PixelStruct Image::pixelBlue = Image::PixelStruct(0,0,1);
const Image::PixelStruct Image::pixelYellow = Image::PixelStruct(1,1,0);
const Image::PixelStruct Image::pixelPink = Image::PixelStruct(1,0,1);
const Image::PixelStruct Image::pixelLiBlue = Image::PixelStruct(0,1,1);


//region CUDA Methods.

/**
 * Displays various properties of all CUDA devices found on machine.
 **/
void display_cuda_device_properties() {
    // Check available CUDA devices.
    cudaDeviceProp cuda_device_properties;
    int cuda_device_count;

    cudaGetDeviceCount(&cuda_device_count);

    for (int index = 0; index < cuda_device_count; index++) {
        cudaGetDeviceProperties(&cuda_device_properties, index);

        printf("\n\n\n\n");
        printf("==== General Information for CUDA Device %d ====\n", index);
        printf("Name: %s\n", cuda_device_properties.name);
        printf("Compute Capabilities: %d.%d\n", cuda_device_properties.major, cuda_device_properties.minor);
        printf("Clock Rate: %d\n", cuda_device_properties.clockRate);
        printf("\n\n");
        printf("==== Memory Information for CUDA Device %d ====\n", index);
        printf("Total Global Memory: %ld\n", cuda_device_properties.totalGlobalMem);
        printf("Total Constant Memory: %ld\n", cuda_device_properties.totalConstMem);
        printf("Max Memory Pitch: %ld\n", cuda_device_properties.memPitch);
        printf("Texture Alignment: %ld\n", cuda_device_properties.textureAlignment);
        printf("\n\n");
        printf("==== Multiprocessing Information for CUDA Device %d ====\n", index);
        printf("Multiprocessor Count: %d\n", cuda_device_properties.multiProcessorCount);
        printf("Shared Memory Per MP: %ld\n", cuda_device_properties.sharedMemPerBlock);
        printf("Registers Per MP: %d\n", cuda_device_properties.regsPerBlock);
        printf("Threads in Warp: %d\n", cuda_device_properties.warpSize);
        printf("Max Threads Per Block: %d\n", cuda_device_properties.maxThreadsPerBlock);
        printf("Max Thread Dimensions: (%d, %d, %d)\n",
            cuda_device_properties.maxThreadsDim[0],
            cuda_device_properties.maxThreadsDim[1],
            cuda_device_properties.maxThreadsDim[2]
        );
        printf("Max Grid Dimensions: (%d, %d, %d)\n",
            cuda_device_properties.maxGridSize[0],
            cuda_device_properties.maxGridSize[1],
            cuda_device_properties.maxGridSize[2]
        );
        printf("\n\n\n\n");
    }
}

/**
 * Updates image so that every other pixel alternates between pure black and pure white.
 * Purely for testing purposes.
 */
// __global__ void cuda_pixel_alternating(Image::PixelStruct* pixel_arr, int pixel_index) {
__global__ void cuda_alternating_pixels(unsigned long total_pixels, unsigned long chunk_size, int chunk_counter, Image::PixelStruct* pixel_arr) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {

        // Manipulate pixel data based on index.
        if (pixel_index % 2 == 0) {
            // Set to black.
            pixel_arr[pixel_index].r = 0.f;
            pixel_arr[pixel_index].g = 0.f;
            pixel_arr[pixel_index].b = 0.f;
        } else {
            // Set to white.
            pixel_arr[pixel_index].r = 255.f;
            pixel_arr[pixel_index].g = 255.f;
            pixel_arr[pixel_index].b = 255.f;
        }
    }
}


/**
 * Updates image so that blocks of pixels alternate between pure black and pure white.
 * Effectively, it's similar to cuda_alternating_pixels, except that all image sizes should look roughly the same.
 * Purely for testing purposes.
 */
// __global__ void cuda_pixel_alternating(Image::PixelStruct* pixel_arr, int pixel_index) {
__global__ void cuda_alternating_blocks(unsigned long total_pixels, unsigned long chunk_size, int chunk_counter, int pixel_block_size, int image_width, int image_height, Image::PixelStruct* pixel_arr) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {

        // Calculate if falls under black or white pixel block.
        int row = pixel_index / image_width;
        int col = pixel_index % image_width;
        int row_grouping = row / pixel_block_size;
        int col_grouping = col / pixel_block_size;
        float color;
        if (row_grouping % 2 == 0) {
            if ((col_grouping % 2) == 0) {
                color = 255.f;
            } else {
                color = 0.f;
            }
        } else {
            if ((col_grouping % 2) == 0) {
                color = 0.f;
            } else {

                color = 255.f;
            }
        }

        // Set pixel color.
        pixel_arr[pixel_index].r = color;
        pixel_arr[pixel_index].g = color;
        pixel_arr[pixel_index].b = color;

        // int color = row * col * 2;
        // printf("Width: %d    Height: %d    Index: %ld    Row: %d    Col: %d    Color: %d\n", image_width, image_height, pixel_index, row, col, color);
        // pixel_arr[pixel_index].r = static_cast<float>(color) / 255.f;
        // pixel_arr[pixel_index].g = static_cast<float>(color) / 255.f;
        // pixel_arr[pixel_index].b = static_cast<float>(color) / 255.f;
    }
}


/**
 * Updates image to increase percieved value of red in image.
 * Purely for testing purposes.
 */
__global__ void cuda_increase_red(unsigned long total_pixels, unsigned long chunk_size, int chunk_counter, Image::PixelStruct* pixel_arr) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {

        // Calculate new pixel colors.
        float red = pixel_arr[pixel_index].r + 0.1f;
        float green = pixel_arr[pixel_index].g - 0.1f;
        float blue = pixel_arr[pixel_index].b - 0.1f;

        // Handle if colors exceed expected values.
        if (red > 1.0f) {
            red = 1.f;
        }
        if (green < 0.0f) {
            green = 0.f;
        }
        if (blue < 0.0f) {
            blue = 0.f;
        }

        // Update pixel values.
        pixel_arr[pixel_index].r = red;
        pixel_arr[pixel_index].g = green;
        pixel_arr[pixel_index].b = blue;
    }
}


/**
 * Updates image to increase percieved value of green in image.
 * Purely for testing purposes.
 */
__global__ void cuda_increase_green(unsigned long total_pixels, unsigned long chunk_size, int chunk_counter, Image::PixelStruct* pixel_arr) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {

        // Calculate new pixel colors.
        float red = pixel_arr[pixel_index].r - 0.1f;
        float green = pixel_arr[pixel_index].g + 0.1f;
        float blue = pixel_arr[pixel_index].b - 0.1f;

        // Handle if colors exceed expected values.
        if (red < 0.0f) {
            red = 0.f;
        }
        if (green > 1.0f) {
            green = 1.f;
        }
        if (blue < 0.0f) {
            blue = 0.f;
        }

        // Update pixel values.
        pixel_arr[pixel_index].r = red;
        pixel_arr[pixel_index].g = green;
        pixel_arr[pixel_index].b = blue;
    }
}


/**
 * Updates image to increase percieved value of blue in image.
 * Purely for testing purposes.
 */
__global__ void cuda_increase_blue(unsigned long total_pixels, unsigned long chunk_size, int chunk_counter, Image::PixelStruct* pixel_arr) {
    // Calculate current index.
    unsigned long pixel_index = blockIdx.x * blockDim.x + threadIdx.x;

    // Adjust index for current chunk.
    pixel_index += (chunk_size * chunk_counter);

    // Only proceed if GPU thread corresponds to a pixel in the image. Ignore all extra threads.
    if (pixel_index <= total_pixels) {

        // Calculate new pixel colors.
        float red = pixel_arr[pixel_index].r - 0.1f;
        float green = pixel_arr[pixel_index].g - 0.1f;
        float blue = pixel_arr[pixel_index].b + 0.1f;

        // Handle if colors exceed expected values.
        if (red < 0.0f) {
            red = 0.f;
        }
        if (green < 0.0f) {
            green = 0.f;
        }
        if (blue > 1.0f) {
            blue = 1.f;
        }

        // Update pixel values.
        pixel_arr[pixel_index].r = red;
        pixel_arr[pixel_index].g = green;
        pixel_arr[pixel_index].b = blue;
    }
}

//endregion CUDA Methods.


//region Constructors.

/**
 * Base constructor.
 */
Image::Image() {
    logger.debug("Image::Image Constructor():");
    logger.debug("");

    // Currently test logic.

    // Essentially creates a 10x10 pixel test image.
    input_path = "../input/pbm/400x300/hallway_00.pbm";
    output_path = "../output/test_00.pbm";
    width = 10;
    height = 10;
    bit_depth = 255;
    PixelStruct c = pixelBlack; // Default color. Can select based on below preset colors.

    // Create and populate image.
    pixel_arr = new PixelStruct[width * height];
    for (int index = 0; index < (width * height); index++) {
        pixel_arr[index] = c;
    }
}


/**
 * Constructor that takes input path to image file.
 * Output path is determined based on parsing input path.
 */
Image::Image(std::string user_path) {
    logger.debug("Image::Image Constructor():");
    logger.debug("    user_path: " + user_path);
    logger.debug("");

    logger.info("Constructing image class.");

    // Re-validate path for saving to class.
    path_validator_struct* return_struct = validate_path(user_path);
    int path_type = return_struct->file_type;
    if (return_struct->err_code) {
        logger.error("Error verifying file type: " + return_struct->err_code.message());
        return;
    }
    free_path_validator_struct(return_struct);
    if (path_type == TYPE_FILE) {
        input_path = user_path;
        calculate_paths(user_path);
        import_image_file();
    } else {
        logger.error("Provided path is not an image file. Class initialization failed.");
        return;
    }

    logger.info("Image class created.");
}


/**
 * Deconstructor.
 */
Image::~Image() {
    logger.debug("Deconstructing image class");

    // Deallocate memory objects.
    delete[] pixel_arr;
}

//endregion Constructors.


//region Private Methods.

/**
 * Calculates class input_path and output_path values.
 *
 * :param image_path: Path to an image on the system.
 */
void Image::calculate_paths(std::string image_path) {
    logger.debug("Image::calculate_paths():");
    logger.debug("    image_path: " + image_path);
    logger.debug("");

    // // Calculate output path location.
    output_path = "../output/";
    std::string base_filename = image_path.substr(image_path.find_last_of("/\\") + 1);

    // Grab directory substrings and generate full output path.
    size_t pos = 0;
    std::string token;
    while ((pos = image_path.find("/")) != std::string::npos) {

        // Pull current substring value from input path.
        token = image_path.substr(0, image_path.find("/"));
        image_path.erase(0, pos + 1);

        // Check if directory value is value we care about.
        if ( (token != ".") && (token != "..") && (token != "input") ) {
            output_path += token + "/";
        }

        // Check if directory exists.
        struct stat info;
        if ( (stat(output_path.c_str(), &info) != 0) || (! info.st_mode & S_IFDIR)) {
            // Directory does not exist.
            mkdir(output_path.c_str(), 0777);
        }

    }

    output_path += base_filename;
}


/**
 * Imports image file data, using class variable data.
 */
void Image::import_image_file() {
    logger.debug("Image::import_image_file():");
    logger.debug("");

    logger.info("Attempting to import image data...");
    logger.info("Data from: \"" + input_path + "\"");

    // Open input file.
    std::ifstream input_file;
    input_file.open(input_path, std::ios::binary | std::ios::in);

    if (! input_file.good()) {
        // Failed to open file.
        logger.error("Failed to open file \"" + input_path + "\"");
    } else {
        // File opened. Grab data.
        std::string header_format;
        input_file >> header_format;

        // Check image data format type.
        if (strcmp(header_format.c_str(), "P6") == 0) {
            // P6 indicates standard "binary RGB format".
            logger.debug("Image is binary RGB format.");

        } else if (strcmp(header_format.c_str(), "P5") == 0) {
            // P5 indicates standard "binary greyscale format".
            logger.debug("Image is binary greyscale format.");

        } else {
            // throw("Can't read input file");
            logger.error("");
            logger.error("Image: \"" + input_path + "\"");
            logger.error("Unhandled image format type: " + header_format);
            return;
        }

        // Save image values to class.
        input_file >> width >> height >> bit_depth;

        // Create and populate image.
        pixel_arr = new PixelStruct[width * height];

        // Skip empty lines in necessary until we get to the binary data.
        input_file.ignore(256, '\n');

        // Read pixels based on image format.
        if (strcmp(header_format.c_str(), "P6") == 0) {
            // Binary RBG format.
            // Read each pixel one by one and convert RGB channel bytes to floats.
            unsigned char pixel[3];
            for (int index = 0; index < width * height; ++index) {
                input_file.read(reinterpret_cast<char *>(pixel), 3);
                pixel_arr[index].r = pixel[0] / 255.f;
                pixel_arr[index].g = pixel[1] / 255.f;
                pixel_arr[index].b = pixel[2] / 255.f;
            }
        } else if (strcmp(header_format.c_str(), "P5") == 0) {
            // Binary RBG format.
            // Read each pixel one by one and convert bytes to floats.
            unsigned char* pixel;
            pixel = (unsigned char*)calloc(1, sizeof(unsigned char));
            for (int index = 0; index < width * height; ++index) {
                input_file.read(reinterpret_cast<char *>(pixel), 1);
                pixel_arr[index].r = *pixel / 255.f;
                pixel_arr[index].g = *pixel / 255.f;
                pixel_arr[index].b = *pixel / 255.f;
            }
            free(pixel);
        }
    }

    input_file.close();
}

//endregion Private Methods.


//region Public Methods.

/**
 * Displays general properties of image held in class.
 */
void Image::display_properties() {
    logger.debug("Image::display_properties():");
    logger.debug("");

    logger.info("");
    logger.info("input file path: " + input_path);
    logger.info("output file path: " + output_path);
    logger.info("width: " + std::to_string(width));
    logger.info("height: " + std::to_string(height));
    logger.info("bit depth: " + std::to_string(bit_depth));
    logger.info("");
}


/**
 * Saves class image data to output file.
 */
void Image::save() {
    logger.debug("Image::save():");
    logger.debug("");

    logger.debug("Attempting to save image to \"" + output_path + "\".");

    // Open output file stream.
    std::ofstream output_file;

    // Attempt to save file data.
    try {
        // Attempt to open file.
        output_file.open(output_path, std::ios::binary | std::ios::out);

        // Check if opened successfully.
        if (output_file.fail()) {
            // Failed to open.
            logger.error("Failed to open file \"" + output_path + "\"");
        } else {
            // Opened successfully.

            // Write image "header" data.
            output_file << "P6\n" << width << " " << height << "\n255\n";

            // Write image pixel data.
            unsigned char r, g, b;
            for (int index = 0; index < (width * height); index++) {
                r = static_cast<float>(std::min(1.f, pixel_arr[index].r) * 255);
                g = static_cast<float>(std::min(1.f, pixel_arr[index].g) * 255);
                b = static_cast<float>(std::min(1.f, pixel_arr[index].b) * 255);
                output_file << r << g << b;
            }

        }

        // Close file stream.
        output_file.close();
        logger.info("Image saved.");

    } catch (const char* err) {
        logger.error("Failed to save image: " + std::string(err));
        output_file.close();
    }
}


/**
 * Method to actually do work on image.
 */
void Image::compute() {
    logger.debug("Image::compute():");
    logger.debug("");

    // Get required CUDA device data.
    // For now, lazily gets first CUDA device and assumes it's the desired device to use.
    // TODO: Handle multiple CUDA devices.
    cudaDeviceProp cuda_device_properties;
    cudaGetDeviceProperties(&cuda_device_properties, 0);
    int cuda_device_max_blocks = cuda_device_properties.maxThreadsDim[0];
    int cuda_device_max_threads = cuda_device_properties.maxThreadsPerBlock;
    unsigned long cuda_chunk_size = cuda_device_max_blocks * cuda_device_max_threads;

    logger.info("CUDA Max Blocks: " + std::to_string(cuda_device_max_blocks));
    logger.info("CUDA Max Threads: " + std::to_string(cuda_device_max_threads));
    logger.info("CUDA Chunk Size: " + std::to_string(cuda_chunk_size));
    logger.info("CUDA Global Mem: " + std::to_string(cuda_device_properties.totalGlobalMem));

    // Calculate some image data.
    int pixel_struct_size = sizeof(PixelStruct);
    unsigned long total_pixels = width * height;
    unsigned long struct_byte_count = total_pixels * pixel_struct_size;

    logger.info("Total Image Size: " + std::to_string(total_pixels));
    logger.info("Pixel Struct Size: " + std::to_string(struct_byte_count));

    // We handle image in chunks, in case we're dealing with large images with more pixels than CUDA device blocks/threads.
    int total_image_chunks = (total_pixels / cuda_chunk_size) + 1;
    logger.info("Total Image Chunks: " + std::to_string(total_image_chunks));
    logger.info("Desired CUDA block count: " + std::to_string(cuda_device_max_blocks));
    logger.info("Desired CUDA thread count: " + std::to_string(cuda_device_max_threads));

    // Process image by iterating over all chunks.
    for (int chunk_index = 0; chunk_index < total_image_chunks; chunk_index++) {
        logger.info("Iterating over image chunk " + std::to_string(chunk_index));
        logger.info("Accounts for pixels " + std::to_string(cuda_chunk_size * chunk_index) + " through " + std::to_string((cuda_chunk_size * chunk_index) + cuda_chunk_size) );

        // Allocate memory on GPU.
        PixelStruct* gpu_pixel_arr;
        cudaMalloc((void**) &gpu_pixel_arr, struct_byte_count);
        cudaMemcpy(gpu_pixel_arr, pixel_arr, struct_byte_count, cudaMemcpyHostToDevice);

        // Run GPU kernel logic.
        // cuda_alternating_pixels<<<cuda_device_max_blocks, cuda_device_max_threads>>>(total_pixels, cuda_chunk_size, chunk_index, gpu_pixel_arr);

        // int pixel_block_size = width / 10;
        // if (pixel_block_size <= 1) {
        //     pixel_block_size = 2;
        // }
        // cuda_alternating_blocks<<<cuda_device_max_blocks, cuda_device_max_threads>>>(total_pixels, cuda_chunk_size, chunk_index, pixel_block_size, width, height, gpu_pixel_arr);

        // cuda_increase_red<<<cuda_device_max_blocks, cuda_device_max_threads>>>(total_pixels, cuda_chunk_size, chunk_index, gpu_pixel_arr);
        // cuda_increase_green<<<cuda_device_max_blocks, cuda_device_max_threads>>>(total_pixels, cuda_chunk_size, chunk_index, gpu_pixel_arr);
        cuda_increase_blue<<<cuda_device_max_blocks, cuda_device_max_threads>>>(total_pixels, cuda_chunk_size, chunk_index, gpu_pixel_arr);

        // Synchronize all GPU computations before iterating over next chunk.
        cudaDeviceSynchronize();

        // Retrieve results.
        cudaMemcpy(pixel_arr, gpu_pixel_arr, struct_byte_count, cudaMemcpyDeviceToHost);
        cudaFree(gpu_pixel_arr);
    }

    logger.debug("Image::compute() finished.");
    logger.debug("");
}

//endregion Public Methods.
